// This data file defines all available message arguments.
// It should only contain calls to DECLARE_MSG_ARG(name, description)

DECLARE_MSG_ARG(error, "")
DECLARE_MSG_ARG(value, "")
DECLARE_MSG_ARG(pretty_value, "")
DECLARE_MSG_ARG(expected, "")
DECLARE_MSG_ARG(actual, "")
DECLARE_MSG_ARG(list, "")
DECLARE_MSG_ARG(old_value, "")
DECLARE_MSG_ARG(new_value, "")

DECLARE_MSG_ARG(action_index, "340")
DECLARE_MSG_ARG(actual_version, "1.3.8")
DECLARE_MSG_ARG(arch, "x64")
DECLARE_MSG_ARG(base_url, "azblob://")
DECLARE_MSG_ARG(binary_source, "azblob")
DECLARE_MSG_ARG(build_result, "One of the BuildResultXxx messages (such as BuildResultSucceeded/SUCCEEDED)")
DECLARE_MSG_ARG(byte_offset, "42")
DECLARE_MSG_ARG(column, "42")
DECLARE_MSG_ARG(command_line, "vcpkg install zlib")
DECLARE_MSG_ARG(command_name, "install")
DECLARE_MSG_ARG(commit_sha, "7cfad47ae9f68b183983090afd6337cd60fd4949")
DECLARE_MSG_ARG(count, "42")
DECLARE_MSG_ARG(cmake_var, "VCPKG_POLICY_EMPTY_PACKAGE")
DECLARE_MSG_ARG(elapsed, "3.532 min")
DECLARE_MSG_ARG(env_var, "VCPKG_DEFAULT_TRIPLET")
DECLARE_MSG_ARG(error_msg, "File Not Found")
DECLARE_MSG_ARG(exit_code, "127")
DECLARE_MSG_ARG(expected_version, "1.3.8")
DECLARE_MSG_ARG(extension, ".exe")
DECLARE_MSG_ARG(feature, "avisynthplus")
DECLARE_MSG_ARG(git_tree_sha, "7cfad47ae9f68b183983090afd6337cd60fd4949")
DECLARE_MSG_ARG(json_field, "identifer")
DECLARE_MSG_ARG(json_type, "an array of identifiers")
DECLARE_MSG_ARG(lower, "42")
DECLARE_MSG_ARG(new_scheme, "version")
DECLARE_MSG_ARG(old_scheme, "version-string")
DECLARE_MSG_ARG(option, "editable")
DECLARE_MSG_ARG(path, "/foo/bar")
DECLARE_MSG_ARG(path_destination, "/foo/bar")
DECLARE_MSG_ARG(path_source, "/foo/bar")
DECLARE_MSG_ARG(row, "42")
DECLARE_MSG_ARG(sha,
                "eb32643dd2164c72b8a660ef52f1e701bb368324ae461e12d70d6a9aefc0c9573387ee2ed3828037ed62bb3e8f566416a2"
                "d3b3827a3928f0bff7c29f7662293e")
DECLARE_MSG_ARG(supports_expression, "windows & !static")
DECLARE_MSG_ARG(system_api, "CreateProcessW")
DECLARE_MSG_ARG(system_name, "Darwin")
DECLARE_MSG_ARG(tool_name, "signtool")
DECLARE_MSG_ARG(triplet, "x64-windows")
DECLARE_MSG_ARG(upper, "42")
DECLARE_MSG_ARG(url, "https://github.com/microsoft/vcpkg")
DECLARE_MSG_ARG(vcpkg_line_info, "/a/b/foo.cpp(13)")
DECLARE_MSG_ARG(vendor, "Azure")
DECLARE_MSG_ARG(version, "1.3.8")

// Choose carefully between the following when describing packages
DECLARE_MSG_ARG(package_name, "zlib")
DECLARE_MSG_ARG(package_abi, "bb1c96759ac96102b4b18215db138daedd3eb16c2cd3302ae7bffab2b643eb87")
DECLARE_MSG_ARG(spec, "zlib:x64-windows")
DECLARE_MSG_ARG(feature_spec, "zlib[featurea,featureb]")
DECLARE_MSG_ARG(version_spec, "zlib:x64-windows@1.0.0")
DECLARE_MSG_ARG(constraint_origin, "zlib:x64-windows@1.0.0")
