#pragma once

namespace vcpkg
{
    struct CoffFileHeader;
    struct CommonPEOptionalHeaders;
    struct UniquePEOptionalHeaders;
    struct UniquePEPlusOptionalHeaders;
    struct ImageDataDirectory;
    struct SectionTableHeader;
    struct ImportDirectoryTableEntry;
    struct ImageLoadConfigCodeIntegrity;
    struct ImageLoadConfigDirectory32;
    struct ImageLoadConfigDirectory64;
    struct DllMetadata;
}